.include  "m48def.inc"    ;definicje dla ATmega48
;.include  "m88def.inc"    ;definicje dla ATmega88
;.include  "m168def.inc"   ;definicje dla ATmega168
;.include  "m328def.inc"   ;definicje dla ATmega328
 
.cseg             ;pami programu (FLASH)
.org      $0000   ;wektory przerwa (dla ATmega48/88: 26 wektorw 2-bajtowych, o adresach $0000-$0019,
                  ;dla ATmega168/328: 26 wektorw 4-bajtowych, o adresach $0000-$0032).

;Jeli adne przerwania nie bd uywane, to w tym miejscu mona bezporednio umieci kod programu

;Wektor nr 1 (Reset)
          ;rjmp    RESET   ;dla ATmega48/88 musi to by instrukcja RJMP o rozmiarze 2 bajtw,
          ;jmp     RESET   ;dla ATmega168/328 musi to by instrukcja JMP o rozmiarze 4 bajtw,
                          ;poniewa wszystkie wektory w tych mikrokontrolerach maj taki rozmiar.

;RESET:
;Inicjalizacja wskanika stosu (zbdna dla ATmega48/88/168/328)
          ;ldi     R16,LOW(RAMEND)
          ;out     SPL,R16
          ;ldi     R16,HIGH(RAMEND)
          ;out     SPH,R16
          
;Konfiguracja mikrokontrolera
          ldi     R16,$01   ;czstotliwo pracy mikrokontrolera (CLK): 1-20 [MHz]
          mov     R15,R16
          ldi     R16,$EF   ;zatrzymanie zbdnych moduw (ADC, TWI, SPI, USART, TC0, TC1, TC2)
          sts     $64,R16   ;zapis rejestru PRR
          ldi     R16,$80   ;wyczenie zasilania komparatora analogowego
          sts     $50,R16   ;zapis rejestru ACSR

;Obsuga wywietlacza LCD 2x16 znakw (HD44780), z wyprowadzeniami podczonymi w nastpujcy sposb:
;R/W=GND, E=PD7, RS=PD6, D7=PD3, D6=PD2, D5=PD1, D4=PD0, D0-D3=NC (nie podczone).
;Po wczeniu zasilania wywietlacz LCD: zostaje wyczyszczony (kursor na pozycji 1 w 1 linii), dziaa
;w trybie 8-bitowym, moe wywietla 1 lini znakw o rozmiarze 5x8 pikseli, wywietlacz, kursor i jego
;miganie s wyczone, kursor porusza si do przodu, przesuwanie (shift) jest wyczone.
          rcall   LCDi      ;inicjalizacja LCD (wczenie trybu 4-bitowego)
          ldi     R18,$28   ;ustawienie opcji: 2 linie, znaki 5x8 pikseli, interfejs 4-bit
          rcall   LCDwi     ;zapis instrukcji do LCD
          ldi     R18,$0C   ;wczenie LCD bez kursora
          rcall   LCDwi     ;zapis instrukcji do LCD
;Instrukcje wywietlacza LCD [czas ich wykonywania]:
; $01 - czyszczenie LCD i powrt kursora na pozycj 1 w 1 linii [1.52 ms]
; $02 - ustawienie kursora na pozycji 1 w 1 linii [1.52 ms]
; $04 - poruszanie kursora do ty, przesuwanie (shift) wyczone [37 us]
; $05 - poruszanie kursora do ty, przesuwanie (shift) wczone [37 us]
; $06 - poruszanie kursora do przodu, przesuwanie (shift) wyczone [37 us]
; $07 - poruszanie kursora do przodu, przesuwanie (shift) wczone [37 us]
; $08 - wyczenie LCD [37 us]
; $0C - wczenie LCD bez kursora [37 us]
; $0E - wczenie LCD z kursorem [37 us]
; $0F - wczenie LCD z migajcym kursorem [37 us]
; $10 - przesunicie kursora w lewo [37 us]
; $14 - przesunicie kursora w prawo [37 us]
; $18 - przesunicie ekranu w lewo [37 us]
; $1C - przesunicie ekranu w prawo [37 us]
; $80 - ustawienie adresu $00 dla DDRAM (kursor na pozycji 1 w 1 linii) [37 us]
; $C0 - ustawienie adresu $40 dla DDRAM (kursor na pozycji 1 w 2 linii) [37 us]

;Obsuga zworki J1
          sbi     PORTC,$05 ;ustawienie pinu PC5 jako wejcia z wysokim stanem (pull-up)
          in      R16,PINC  ;odczyt stanu pinw PC0-PC6
          andi    R16,$20
          brne    Boot1     ;skok, jeli na pinie PC5 jest stan wysoki (zworka J1=OFF)
;Zworka J1=ON (zwarta)
          rcall   LCDclr    ;czyszczenie LCD
          rcall   FuseBit   ;odczyt i wywietlenie na LCD wartoci fuse/lock bitw
Boot0:    in      R16,PINC  ;odczyt stanu pinw PC0-PC6
          andi    R16,$20
          breq    Boot0     ;skok, jeli na pinie PC5 jest stan niski (zworka J1=ON)
;Zworka J1=OFF (rozwarta)
Boot1:    rcall   LCDclr    ;czyszczenie LCD
          ldi     ZL,LOW(Text1<<1)
          ldi     ZH,HIGH(Text1<<1)
          rcall   LCDwf     ;wywietlenie na LCD cigu ASCIIZ z pamici FLASH
          ldi     R18,$C0   ;ustawienie kursora na pozycji 1 w 2 linii
          rcall   LCDwi     ;zapis instrukcji do LCD
          rcall   LCDwf     ;wywietlenie na LCD cigu ASCIIZ z pamici FLASH

;Zapisywanie 8 znakw uytkownika do pamici CGRAM wywietlacza LCD
          ldi     ZL,LOW(Chars<<1)
          ldi     ZH,HIGH(Chars<<1) ;adres danych opisujcych budow 8 znakw uytkownika [64-bajty]
          ldi     R18,$40   ;ustawienie adresu $00 dla CGRAM (pierwszy znak)
          rcall   LCDwi     ;zapis instrukcji do LCD
          ldi     R22,$40
LCDwfch:  lpm     R18,Z+
          rcall   LCDwc     ;zapis znaku do LCD
          dec     R22
          brne    LCDwfch
;Wywietlanie znaku uytkownika na LCD
          ldi     R18,$CF   ;ustawienie kursora na pozycji 16 w 2 linii
          rcall   LCDwi     ;zapis instrukcji do LCD
          ldi     R18,$05   ;znak uytkownika nr 5
          rcall   LCDwc     ;wywietlenie znaku na LCD

;Inicjalizacja rejestrw
          clr     XL
          clr     XH
          movw    R22,XL
          movw    R24,XL
          inc     XL

;Gwna ptla programu
Loop:     rcall   Add3216       ;dodawanie sowa do wartoci 32-bitowej
          ldi     ZL,LOW(sbuf)
          ldi     ZH,HIGH(sbuf) ;adres bufora wyjciowego w pamici SRAM
          mov     R16,R22       ;najstarszy bajt 32-bitowej wartoci
          rcall   ConBH         ;konwersja bajtu do systemu HEX-ASCIIZ
          mov     R16,R23
          rcall   ConBH         ;konwersja bajtu do systemu HEX-ASCIIZ
          mov     R16,R24
          rcall   ConBH         ;konwersja bajtu do systemu HEX-ASCIIZ
          mov     R16,R25       ;najmodszy bajt 32-bitowej wartoci
          rcall   ConBH         ;konwersja bajtu do systemu HEX-ASCIIZ
          sbiw    ZL,$08
          ldi     R18,$C6       ;ustawienie kursora na pozycji 7 w 2 linii
          rcall   LCDwi         ;zapis instrukcji do LCD
          rcall   LCDws         ;wywietlenie na LCD cigu ASCIIZ z pamici SRAM
          rjmp    Loop

;Dodawanie sowa do wartoci 32-bitowej: [R22:R23:R24:R25] + [XH:XL] = [R22:R23:R24:R25]
Add3216:  clr     R2
          add     R25,XL  ;najmodszy bajt wartoci/wyniku, modszy bajt sowa
          adc     R24,XH  ;starszy bajt sowa
          adc     R23,R2
          adc     R22,R2  ;najstarszy bajt wartoci/wyniku
          ret             ;C=1 jeli wynik >4294967295

;Zapis szesnastkowej wartoci bajtu do pamici SRAM w postaci cigu ASCIIZ: "NN"
ConBH:    ldi     R17,$02
ConBH0:   swap    R16     ;bajt do konwersji
          mov     R18,R16
          andi    R18,$0F
          subi    R18,$D0
          cpi     R18,$3A
          brlt    ConBH1
          subi    R18,$F9 ;subi R18,$D9 - mae litery (a-f)
ConBH1:   st      Z+,R18
          dec     R17
          brne    ConBH0
          st      Z,R17   ;wyjcie danych
          ret

;Zapis szesnastkowych wartoci fuse/lock bitw do pamici SRAM w postaci
;cigu ASCIIZ: "FL:LB:FE:FH" i wywietlenie go na wywietlaczu LCD.
FuseBit:  ldi     YL,LOW(sbuf)
          ldi     YH,HIGH(sbuf) ;adres wyjcia danych
          clr     ZL
          clr     ZH
          ldi     R18,$3A
FuseBit0: lds     R16,$57       ;odczyt rejestru SPMCSR
          ori     R16,$09       ;ustawienie bitw BLBSET i SELFPRGEN
          sts     $57,R16       ;zapis rejestru SPMCSR
          lpm     R2,Z+         ;odczyt fuse/lock bitw
;W zalenoci od wartoci adresu [ZH:ZL] s odczytywane:
; $0000 - Fuse bits Low byte (FL)
; $0001 - Lock Bits (LB)
; $0002 - Fuse bits Extended byte (FE)
; $0003 - Fuse bits High byte (FH)
          ldi     R16,$02
FuseBit1: swap    R2
          mov     R17,R2
          andi    R17,$0F
          subi    R17,$D0
          cpi     R17,$3A
          brlt    FuseBit2
          subi    R17,$F9
FuseBit2: st      Y+,R17
          dec     R16
          brne    FuseBit1
          st      Y+,R18
          cpi     ZL,$04
          brne    FuseBit0
          st      -Y,R16
;Wywietlanie wartoci fuse/lock bitw na LCD
          ldi     ZL,LOW(Text0<<1)
          ldi     ZH,HIGH(Text0<<1)
          rcall   LCDwf         ;wywietlenie na LCD cigu ASCIIZ z pamici FLASH
          ldi     R18,$C0       ;ustawienie kursora na pozycji 1 w 2 linii
          rcall   LCDwi         ;zapis instrukcji do LCD
          ldi     ZL,LOW(sbuf)
          ldi     ZH,HIGH(sbuf)
          rcall   LCDws         ;wywietlenie na LCD cigu ASCIIZ z pamici SRAM
          ret

;Inicjalizacja wywietlacza LCD
LCDi:     ldi     R16,$CF
          out     DDRD,R16  ;ustawienie pinw PD0-PD3/PD6-PD7 jako wyj, PD4-PD5 jako wej
          clr     R16       ;E=0, RS=0
          out     PORTD,R16 ;wyzerowanie wyj PD0-PD3/PD6-PD7
;Czekanie na ustabilizowanie si napicia zasilania
          ldi     R16,$C8   ;opnienie 20 ms
          rcall   Wait      ;ptla opniajca
          rcall   Wait      ;ptla opniajca
;Wysanie trzech instrukcji $30 do LCD
          ldi     R18,$03
LCDi0:    ldi     R16,$83
          out     PORTD,R16 ;E=1, RS=0
          ;(wait 4 CL)
          cbi     PORTD,$07 ;E=0, RS=0
          ldi     R16,$32   ;opnienie 5 ms
          rcall   Wait      ;ptla opniajca
          dec     R18
          brne    LCDi0
;Wysanie instrukcji $20 do LCD (wczenie trybu 4-bitowego)
          ldi     R16,$82
          out     PORTD,R16 ;E=1, RS=0
          ;(wait 4 CL)
          cbi     PORTD,$07 ;E=0, RS=0
          ldi     R16,$0A   ;opnienie 1 ms
          rcall   Wait      ;ptla opniajca
          ret

;Zapis instrukcji do wywietlacza LCD (tryb 4-bitowy)
LCDwi:    ldi     R20,$80   ;E=1, RS=0
          rjmp    LCDwc0
;Zapis znaku do wywietlacza LCD (tryb 4-bitowy)
LCDwc:    ldi     R20,$C0   ;E=1, RS=1
LCDwc0:   ldi     R21,$02
LCDwc1:   swap    R18       ;bajt do wysania
          mov     R19,R18
          andi    R19,$0F
          andi    R20,$F0   ;E=1, RS=pocztkowa warto
          add     R20,R19
          out     PORTD,R20 ;wysanie 4-bitw do LCD (najpierw starsze, pniej modsze)
          ;(wait 4 CL)
          cbi     PORTD,$07 ;E=0, RS=pocztkowa warto
          ldi     R16,$01   ;opnienie 100 us
          rcall   Wait      ;ptla opniajca
          dec     R21
          brne    LCDwc1
          ret

;Czyszczenie wywietlacza LCD i powrt kursora na pozycj 1 w 1 linii
LCDclr:   ldi     R18,$01 ;czyszczenie LCD i powrt kursora
          rcall   LCDwi   ;zapis instrukcji do LCD
          ldi     R16,$10 ;opnienie 1.6 ms
          rcall   Wait    ;ptla opniajca
          ret

;Zapis cigu ASCIIZ z pamici SRAM do wywietlacza LCD (tryb 4-bitowy)
LCDws0:   rcall   LCDwc   ;zapis znaku do LCD
LCDws:    ld      R18,Z+  ;odczyt znaku z pamici SRAM
          tst     R18
          brne    LCDws0
          ret

;Zapis cigu ASCIIZ z pamici FLASH do wywietlacza LCD (tryb 4-bitowy)
LCDwf0:   rcall   LCDwc   ;zapis znaku do LCD
LCDwf:    lpm     R18,Z+  ;odczyt znaku z pamici FLASH
          tst     R18
          brne    LCDwf0
          ret

;Ptla opniajca o zadany czas: 100u - 25.5m [s]
Wait:     mov     R2,R15  ;R15 (CLK) = czstotliwo pracy mikrokontrolera: $01-$14 (1-20 MHz)
          ldi     R17,$19
          mul     R16,R17 ;R16 (DLY) = warto opnienia: $01-$FF (100us-25.5ms)
Wait0:    movw    YL,R0
Wait1:    sbiw    YL,$01
          brne    Wait1
          dec     R2
          brne    Wait0
          ret

;Napisy dla wywietlacza LCD
Text0:    .db     "Fuse/lock bits:",0
Text1:    .db     "*** LCD test ***",0,"Count:",0

;Znaki uytkownika dla wywietlacza LCD
Chars:    .db     17,14,10,10,10,14,17,0 ;0
          .db     17,4,12,4,4,14,17,0    ;1
          .db     17,14,2,14,8,14,17,0   ;2
          .db     17,14,2,6,2,14,17,0    ;3
          .db     17,10,10,14,2,2,17,0   ;4
          .db     17,14,8,14,2,14,17,0   ;5
          .db     17,14,8,14,10,14,17,0  ;6
          .db     17,14,10,2,2,2,17,0    ;7

.dseg                ;pami danych (SRAM)
.org      $0100
sbuf:     .byte   12 ;rezerwuje N-bajtw w pamici SRAM